//
//  OCEPathObject.m
//  InfiniteCanvas
//
//  Created by Markus Mueller <markus@mindnode.com>.
//	http://www.infinitecanvas.com/
//  Copyright (c) 2003-2005 Markus Mueller. All rights reserved.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//  

#import "OCEPathObject.h"
#import "OCERectObject.h"
#import "OCECanvas.h"
#import "NSBezierPathHitTest.h"
#import "OCEMath.h"
#import "OCEDocument.h"
#import "OCETrailObject.h"
#import "OCENavTrailObject.h"

@implementation OCEPathObject
- (id)init 
{
    [super init];
    path = nil;
    color = [[NSColor blackColor] retain];
    thickness = 3;
    startRect = nil;
    endRect = nil;
	arrowRotation1 = 0; // endPointRotation
    arrowRotation2 = 0; // startPointRotation
	arrow = 1;
	allowFreeEndPoints = YES;
	
    return self;
}

- (void)dealloc 
{
    if (path)
        [path release];
    [NSColor release];
    [super dealloc];
}

// ----------------------------------------------------------------------
// Create
// ----------------------------------------------------------------------
- (id) initWithPoint: (NSPoint) point
{
    [self init];
    path = [[NSBezierPath bezierPath] retain];
    [path moveToPoint: point];
	[path setLineWidth:thickness];
	return self;
}

- (void) addLineToPoint: (NSPoint) point
{
    [path lineToPoint: point];
	[self calculateArrowRotation];
}

- (void) addCurveToPoint: (NSPoint) point controlPoint1: (NSPoint) controlPoint1 controlPoint2: (NSPoint) controlPoint2
{
    [path curveToPoint: point controlPoint1: controlPoint1 controlPoint2: controlPoint2];
	[self calculateArrowRotation];
}

// ----------------------------------------------------------------------
// Draw
// ----------------------------------------------------------------------
- (void) drawObjectWithFraction: (float) fraction isSelected: (bool) selected
{	
	if (fraction == 1)
		[color set];
	else
		[[color colorWithAlphaComponent:fraction] set];
	
	
	[path stroke];
	
	if ((arrow) && ([self valid]))
		[self drawArrows];
    
    if (selected)
        [self drawKnobs];
	
}


- (void) drawKnobs
{
    float length = 6;
    float halfLength = length / 2.0;
    NSPoint *points = malloc(3*sizeof(NSPoint));
    NSBezierPathElement element;
    NSBezierPath *knobPath = [NSBezierPath bezierPath];
    NSBezierPath *controlPath = [NSBezierPath bezierPath];
    NSRect rect;
	NSPoint lastPoint = NSZeroPoint;
    
    // 1. shadow
    int i, c = [path elementCount];
    for (i=0;i<c;i++) 
    {
        element=[path elementAtIndex:i associatedPoints:points];
        switch (element) 
        {
            case NSCurveToBezierPathElement:
                // zuerst die 2 controll punkte
				// first knob
				rect = NSMakeRect(points[0].x-halfLength, points[0].y-halfLength, length, length);
				[knobPath appendBezierPathWithRect: rect];
				// then line from endPoint to knob
				[controlPath moveToPoint:points[0]];
				[controlPath lineToPoint:lastPoint];

				// second knob
				rect = NSMakeRect(points[1].x-halfLength, points[1].y-halfLength, length, length);
				[knobPath appendBezierPathWithRect: rect];
				// then line from endPoint to knob
				[controlPath moveToPoint:points[1]];
				[controlPath lineToPoint:points[2]];
				
				rect = NSMakeRect(points[2].x-halfLength, points[2].y-halfLength, length, length);
                [knobPath appendBezierPathWithRect: rect];
				lastPoint = points[2];
				break;
            case NSLineToBezierPathElement:
            case NSMoveToBezierPathElement:
				// beide besitzen nur einen punkt
                rect = NSMakeRect(points[0].x-halfLength, points[0].y-halfLength, length, length);
                [knobPath appendBezierPathWithRect: rect];
				lastPoint = points[0];
                break;
            default:
                break;
        }
    }
    
    // draw Line to ControlPoints
    [[NSColor grayColor] set];
    [controlPath setLineWidth:0.5];
    [controlPath stroke];
    
    // draw Shadow
    [[NSColor blackColor] set];
    [knobPath fill];
    
    // 2. knob
    NSAffineTransform *knobTransform = [NSAffineTransform transform];
    [knobTransform translateXBy: -1 yBy: 1];
    [knobPath transformUsingAffineTransform:knobTransform];
    
    if (locked)
        [[NSColor redColor] set];
    else
        [[NSColor knobColor] set];
    
    [knobPath fill];
    
    free(points);
}

// ----------------------------------------------------------------------
// Arrow
// ----------------------------------------------------------------------
- (void) drawArrows
{
	// only simple end arrow
	NSAffineTransform *transform;
	NSPoint point;
	NSGraphicsContext *currentContext = [NSGraphicsContext currentContext];
	
	point = [self endPoint];
	transform = [NSAffineTransform transform];
	[transform translateXBy: point.x yBy: point.y];
	[transform rotateByDegrees: arrowRotation1];
	[transform translateXBy: -point.x yBy: -point.y];
			
	NSBezierPath *arrowPath = [NSBezierPath bezierPath];                  
	[arrowPath moveToPoint:NSMakePoint(point.x-10,point.y+10)];
	[arrowPath lineToPoint:point];
	[arrowPath lineToPoint:NSMakePoint(point.x-10,point.y-10)];
	[arrowPath setLineWidth:thickness];			
			
	[currentContext saveGraphicsState];
	[transform concat];
	[arrowPath stroke];
	[currentContext restoreGraphicsState];		
}

- (void) calculateArrowRotation
{
	// we love this method. it is called everytime a point changed. so we can do some checking here
	// check if endPoint != startPoint when there is only a line
	
	
    int count = [path elementCount];
    if (count < 2)
        return;
    NSPoint fromPoint, toPoint;
    
    // end point
    toPoint = [self endPoint];
	
	NSPoint *points = malloc(3*sizeof(NSPoint));
	NSBezierPathElement element = [path elementAtIndex: count-1 associatedPoints:points];
	if (element == NSCurveToBezierPathElement)
		fromPoint = points[1];
	else
	{
		element = [path elementAtIndex: count-2 associatedPoints:points];
		if (element == NSCurveToBezierPathElement)
			fromPoint = points[2];
		else
			fromPoint = points[0];
	}
    
	arrowRotation1 = [OCEMath calculateDegreesVector1: NSMakePoint(1,0) 
											  vector2: NSMakePoint(toPoint.x-fromPoint.x,toPoint.y-fromPoint.y)];
    // start point
    toPoint = [self startPoint];
    
	[path elementAtIndex: 1 associatedPoints:points];
	fromPoint = points[0];
	free(points);
	
	arrowRotation2 = [OCEMath calculateDegreesVector1: NSMakePoint(1,0) 
											  vector2: NSMakePoint(toPoint.x-fromPoint.x,toPoint.y-fromPoint.y)];
}


// ----------------------------------------------------------------------
// Hit Tests and Knobs
// ----------------------------------------------------------------------
- (bool) intersectsRect: (NSRect) rect
{
    return [path intersectsRect:rect];
}

- (int) hitTest: (NSPoint)aPoint isSelected: (bool) isSelected
{
	// 0: not found
	// 1: trail hit
	// 2: point 0 controlPoint 1
	// 3: point 0 controlPoint 2
	// 4: point 0 endPoint
	// 5: point 1 controlPoint 1
	// 6: ...
    if (isSelected)
    {
        float length = 6;
        float halfLength = length / 2.0;
        NSPoint *points = malloc(3*sizeof(NSPoint));
        NSBezierPathElement element;
        NSRect rect;
        
        int i, c = [path elementCount];
        for (i=0;i<c;i++) 
        {
            element=[path elementAtIndex:i associatedPoints:points];
            switch (element) 
            {
                case NSCurveToBezierPathElement:
                    // zuerst die 2 controll punkte
					// first knob
					rect = NSMakeRect(points[0].x-halfLength, points[0].y-halfLength, length, length);
					if (NSMouseInRect(aPoint, rect, NO))
					{
						free(points);
						return i*3+0+2; 
					}
					// second knob
					rect = NSMakeRect(points[1].x-halfLength, points[1].y-halfLength, length, length);
					if (NSMouseInRect(aPoint, rect, NO))
					{
						free(points);
						return i*3+1+2; 
					}
					// end point
					rect = NSMakeRect(points[2].x-halfLength, points[2].y-halfLength, length, length);
                    if (NSMouseInRect(aPoint, rect, NO))
                    {
                        free(points);
                        return i*3+2+2; 
                    }
						
					break;
                case NSLineToBezierPathElement:
                case NSMoveToBezierPathElement:
                    // auch die vorangegangenen elemente besetzen nur einen point => wird gleich behandelt
                    rect = NSMakeRect(points[0].x-halfLength, points[0].y-halfLength, length, length);
                    if (NSMouseInRect(aPoint, rect, NO))
                    {
                        free(points);
                        return i*3+2+2; 
                    }
					break;
                default:
                    break;
            }
        }
		free(points);
    } // selected
    
    /// hit trail
	if ([path isStrokeHitByPoint:aPoint padding:0.5])
		return 1;
	
	return 0;
}

- (enum corner) movingKnob: (enum corner) knob fromPoint: (NSPoint) lastPoint toPoint: (NSPoint) point 
				 withEvent: (NSEvent *) theEvent
{
	if (knob < 2)
		return knob;
	
	if (locked)
		return knob;
	
	knob -= 2;
	int knob_index = knob / 3;// what point has been hit
	int knob_point = knob % 3;// what point at that index

	if ((startRect) && (!knob_index))
		[self unregisterAssociatedStartRect];
	
	if ((endRect) && (knob_point == 2) && (knob_index == [path elementCount]-1))
		[self unregisterAssociatedEndRect];
	
	NSPoint *points = malloc(3*sizeof(NSPoint));
	NSBezierPathElement element = [path elementAtIndex:knob_index associatedPoints:points];
	
	if ((element == NSMoveToBezierPathElement) || (element == NSLineToBezierPathElement))
		points[0] = point;
	else // curveElement
		points[knob_point] = point;
		
	[path setAssociatedPoints: points atIndex: knob_index];
	free(points);
	
	[self calculateArrowRotation];
	[self adaptPathToRects];
	
    return knob+2;
}

- (void) doubleKlickAtPoint: (NSPoint) aPoint
{
	if ((locked) && !(([undoManager isUndoing]) || ([undoManager isRedoing])))
        return;

	[self changeSegment: [path segmentHitByPoint: aPoint]];
}

- (NSPoint) pointForKnob: (enum corner) knob
{
	knob -= 2;
	int knob_index = knob / 3; // what point has been hit
	int knob_point = knob % 3; // what point at that index				
			
	NSPoint *points = malloc(3*sizeof(NSPoint));
	NSPoint returnPoint;
	NSBezierPathElement element=[path elementAtIndex:knob_index associatedPoints:points];
	switch (element) 
	{
		case NSCurveToBezierPathElement:
			// controlPoints mssen nicht an das grid angepasst werden!
			returnPoint = points[knob_point];
			break;
		case NSLineToBezierPathElement:
		case NSMoveToBezierPathElement:
			returnPoint = points[0];
			break;
		default:
			break;
	}
	free(points);
	
	return returnPoint;
}

// ----------------------------------------------------------------------
// Attributes
// ----------------------------------------------------------------------
- (int) arrow
{
    return arrow;
}

- (void) setArrow: (int) newArrow
{
    if (arrow != newArrow)
    {
        [[undoManager prepareWithInvocationTarget:self] setArrow: arrow];
		[undoManager setActionName: @"Change Arrow Type"];
		
        arrow = newArrow;
    }
}

- (NSBezierPath *) path
{
	return path;
}

- (void) startManipulationWithSelector:(SEL)aSelector andData: (id) data;
{
	if ([NSStringFromSelector(aSelector) isEqualToString: @"moveXBy:moveYBy:"])
	{
		NSPoint origin = [self origin];
		NSArray *array = [NSArray arrayWithObjects: [NSNumber numberWithFloat:origin.x], 
													[NSNumber numberWithFloat:origin.y],nil];
		[super startManipulationWithSelector: aSelector andData: array];
	}
	else if ([NSStringFromSelector(aSelector) isEqualToString: @"movingKnob:fromPoint:toPoint:withEvent:"])
	{
		NSPoint *points = malloc(3*sizeof(NSPoint));
		int knob = [data intValue]-2;
		int knob_index = knob / 3;// what point has been hit
		int knob_point = knob % 3;// what point at that index				

		NSBezierPathElement element = [path elementAtIndex:knob_index associatedPoints:points];
		NSPoint point;
		if ((element == NSMoveToBezierPathElement) || (element == NSLineToBezierPathElement))
			point = points[0];
		else
			point = points[knob_point];
		free(points);
		NSArray *array = [NSArray arrayWithObjects: data,[NSNumber numberWithFloat:point.x], 
							[NSNumber numberWithFloat:point.y], startRect, endRect,nil];
		[super startManipulationWithSelector: aSelector andData:array];
	}
	else
	{
		[super startManipulationWithSelector: aSelector andData:data];
	}
}

- (void) stopManipulation
{
	if (!editingObject)
		return;
	
	if ([NSStringFromSelector(manipulationSelector) isEqualToString: @"moveXBy:moveYBy:"])
	{
		NSPoint origin = [self origin];
		float xOffset = origin.x - [[manipulationData objectAtIndex:0] floatValue];
		float yOffset = origin.y - [[manipulationData objectAtIndex:1] floatValue];
		if ((xOffset != 0) || (yOffset != 0))
		{
			[[undoManager prepareWithInvocationTarget:self] moveXBy: -xOffset moveYBy: -yOffset];
			[undoManager setActionName: @"Move"];
		}
	}
	else if ([NSStringFromSelector(manipulationSelector) isEqualToString: @"movingKnob:fromPoint:toPoint:withEvent:"])
	{
		int knob = [[manipulationData objectAtIndex:0] intValue]-2;
		int knob_index = knob / 3;// what point has been hit
		int knob_point = knob % 3;// what point at that index				
		
		if (!allowFreeEndPoints)
		{
			if ((!startRect) && (knob_index == 0))
			{
				[self setStartPoint:NSMakePoint([[manipulationData objectAtIndex:1] floatValue],
												[[manipulationData objectAtIndex:2] floatValue])];
				[self setAssociatedStartRect: [manipulationData objectAtIndex:3]];
				[super stopManipulation];
				[undoManager setActionName: @"Reshape"];
				return;
			}
			else if ((!endRect) && (knob_point == 2) && (knob_index == [path elementCount]-1))
			{
				[self setEndPoint:NSMakePoint([[manipulationData objectAtIndex:1] floatValue],
											  [[manipulationData objectAtIndex:2] floatValue])];
				[self setAssociatedEndRect: [manipulationData objectAtIndex:4]];
				[super stopManipulation];
				[undoManager setActionName: @"Reshape"];
				return;
			}
			
		}
		
		[[undoManager prepareWithInvocationTarget:self] setKnob: knob+2 
							toPoint: NSMakePoint([[manipulationData objectAtIndex:1] floatValue],
												 [[manipulationData objectAtIndex:2] floatValue])];
		[undoManager setActionName: @"Reshape"];
	}
    [super stopManipulation];
}

- (void) gridObject: (int) gridSize
{
	NSPoint *points = malloc(3*sizeof(NSPoint));
    NSBezierPathElement element;    
    int i = 0; 
	int c = [path elementCount];
	if (startRect)
		i = 1;
	
	if (endRect)
		c -= 1;
	
    for (;i<c;i++) 
    {
        element=[path elementAtIndex:i associatedPoints:points];
        switch (element) 
        {
            case NSCurveToBezierPathElement:
                // controlPoints mssen nicht an das grid angepasst werden!
				points[2] = [OCEObject gridPoint: points[2] withGridSize: gridSize];
				break;
            case NSLineToBezierPathElement:
            case NSMoveToBezierPathElement:
				points[0] = [OCEObject gridPoint: points[0] withGridSize: gridSize];
                break;
            default:
                break;
        }
		[path setAssociatedPoints: points atIndex: i];
    }
	free(points);
}

- (void) moveXBy: (double) moveX moveYBy: (double) moveY
{

	if (([undoManager isUndoing]) || ([undoManager isRedoing]))
		[[undoManager prepareWithInvocationTarget:self] moveXBy: -moveX moveYBy: -moveY];
	else
	{
		if (locked)
			return; // don't return when undoing
	}

    NSAffineTransform *transform = [NSAffineTransform transform];
    [transform translateXBy: moveX yBy: moveY];
    [path transformUsingAffineTransform:transform];
	
	[self adaptPathToRects];
}

- (NSPoint) origin
{
    NSPoint points = NSZeroPoint;
    [path elementAtIndex: 0 associatedPoints: &points];
    return points;   
}

- (NSColor *) color
{
	return color;
}

- (void) setColor: (NSColor *) newColor
{
    [[undoManager prepareWithInvocationTarget:self] setColor:color];
	[undoManager setActionName: @"Change Trail Color"];
    
    id old = color;
    color = [newColor retain];
    [old release];
}

- (float) thickness
{
	return thickness;
}

- (void) setThickness: (float) newThickness
{
    [[undoManager prepareWithInvocationTarget:self] setThickness: thickness];
	[undoManager setActionName: @"Change Thickness"];
    
    thickness = newThickness;  
	[path setLineWidth:thickness];
}

- (NSRect) bounds
{
	NSRect myBounds = [path bounds];   
	return (NSIsEmptyRect(myBounds)) ? NSMakeRect(myBounds.origin.x,myBounds.origin.y,1,1) : myBounds;
		
}

- (void) setKnob: (enum corner) knob toPoint: (NSPoint) point
{
	knob -= 2;
	int knob_index = knob / 3;// what point has been hit
	int knob_point = knob % 3;// what point at that index				
		
	NSPoint *points = malloc(3*sizeof(NSPoint));
	NSBezierPathElement element = [path elementAtIndex:knob_index associatedPoints:points];
	if (element == NSCurveToBezierPathElement)
	{
		[[undoManager prepareWithInvocationTarget:self] setKnob: knob+2 toPoint:points[knob_point]];
		points[knob_point] = point;
	}
	else
	{
		[[undoManager prepareWithInvocationTarget:self] setKnob: knob+2 toPoint:points[0]];
		points[0] = point;
	}
	
	[path setAssociatedPoints: points atIndex: knob_index];
	free(points);
	
	[self calculateArrowRotation];
}

- (NSPoint) startPoint
{
	return [self origin];
}

- (void) setStartPoint: (NSPoint) point
{
	
	NSPoint *points = malloc(3*sizeof(NSPoint));
	
	[path elementAtIndex:0 associatedPoints:points];
	if (!editingObject)
		[[undoManager prepareWithInvocationTarget:self] setStartPoint: points[0]];
	points[0] = point;
	[path setAssociatedPoints: points atIndex: 0];

	free(points);
	[self calculateArrowRotation];
}

- (NSPoint) endPoint
{
	NSPoint *points = malloc(3*sizeof(NSPoint));
	int index = [path elementCount]-1;
	NSPoint point;
	NSBezierPathElement element = [path elementAtIndex: index associatedPoints:points];
	if (element == NSCurveToBezierPathElement)
		point = points[2];
	else
		point = points[0];

	free(points);
	
	return point;
}

- (void) setEndPoint: (NSPoint) point
{
	NSPoint *points = malloc(3*sizeof(NSPoint));
	int index = [path elementCount]-1;
	NSBezierPathElement element = [path elementAtIndex: index associatedPoints:points];
	if (element == NSCurveToBezierPathElement)
	{
		if (!editingObject)
			[[undoManager prepareWithInvocationTarget:self] setEndPoint: points[2]];
		points[2] = point;
	}
	else
	{
		if (!editingObject)
			[[undoManager prepareWithInvocationTarget:self] setEndPoint: points[0]];
		points[0] = point;
	}
	[path setAssociatedPoints: points atIndex: index];
	
	[self calculateArrowRotation];
	free(points);
}

- (bool) endPointAtKnob: (int) knob;
{
	knob -=2;
	int knob_index = knob / 3;// what point has been hit
	int knob_point = knob % 3;// what point at that index				
	if ( ([path elementCount]-1 == knob_index) && (knob_point == 2))
	{
		return YES;
	}
	
	return NO;
}


- (void) changeSegment: (int) segment
{
	[[undoManager prepareWithInvocationTarget:self] changeSegment: segment];
	[undoManager setActionName: @"Reshape"];
		
	NSPoint *points = malloc(3*sizeof(NSPoint));
    NSBezierPath *newPath = [[NSBezierPath bezierPath] retain];
	NSBezierPathElement element;
	int i, c = [path elementCount];
	NSPoint lastPoint = NSZeroPoint;
	
	for (i=0;i<c;i++) 
	{
		element=[path elementAtIndex:i associatedPoints:points];
		switch (element) 
		{
			case NSCurveToBezierPathElement:
			{
				if (segment == i)
					[newPath lineToPoint:points[2]];
				else
					[newPath curveToPoint:points[2] controlPoint1: points[0] controlPoint2: points[1]];
				lastPoint = points[2];
				break;
			}
			case NSLineToBezierPathElement:
			{
				if (segment == i)
					[newPath curveToPoint:points[0] 
							controlPoint1: NSMakePoint(lastPoint.x+40,lastPoint.y-40) 
							controlPoint2: NSMakePoint(points[0].x-40,points[0].y+40)];
				else
					[newPath lineToPoint:points[0]];
				lastPoint = points[0];
				break;
			}
			case NSMoveToBezierPathElement:
			{
				[newPath moveToPoint:points[0]];
				lastPoint = points[0];
				break;
			}
			default:
				break;
		}
	}
	
	free(points);
	[path release];
	path = newPath;
	
	[path setLineWidth:thickness];
	[self calculateArrowRotation];
	[self adaptPathToRects];	
}

- (bool) valid
{
	int count = [path elementCount];
	if (count > 1)
		return (((count == 2) && (NSEqualPoints([self startPoint],[self endPoint]))) ? NO : YES);
	else
		return NO;
}

// ----------------------------------------------------------------------
// NSCopying Protocol
// ----------------------------------------------------------------------
- (id)copyWithZone:(NSZone *)zone
{
    OCEPathObject *copy = [super copyWithZone: zone];
    
    copy->path = [path copy];
    copy->color = [color retain];
    copy->thickness = thickness;
    copy->startRect = nil;
    copy->endRect = nil;
    copy->arrowRotation1 = arrowRotation1;
	copy->arrowRotation2 = arrowRotation2;
	copy->arrow = arrow;
	copy->allowFreeEndPoints = allowFreeEndPoints;
    
    return copy;
}



// ----------------------------------------------------------------------
// Save and Load
// ----------------------------------------------------------------------
- (NSMutableArray *)dataRepresentation
{
	NSMutableArray *array = [[super dataRepresentation] retain];
	
	// PATH
	NSMutableArray *pathArray = [NSMutableArray arrayWithCapacity:0];
	
	NSPoint *points = malloc(3*sizeof(NSPoint));
	NSBezierPathElement element;
	
	int i, c = [path elementCount];
	for (i=0;i<c;i++) 
	{
		element=[path elementAtIndex:i associatedPoints:points];
		switch (element) 
		{
			case NSCurveToBezierPathElement:
				[pathArray addObject:[NSNumber numberWithInt: 0]];
				[pathArray addObject:[NSNumber numberWithFloat: points[0].x]];
				[pathArray addObject:[NSNumber numberWithFloat: points[0].y]];
				[pathArray addObject:[NSNumber numberWithFloat: points[1].x]];
				[pathArray addObject:[NSNumber numberWithFloat: points[1].y]];
				[pathArray addObject:[NSNumber numberWithFloat: points[2].x]];
				[pathArray addObject:[NSNumber numberWithFloat: points[2].y]];
					break;
			case NSLineToBezierPathElement:
			case NSMoveToBezierPathElement:
				[pathArray addObject:[NSNumber numberWithInt: 1]];
				[pathArray addObject:[NSNumber numberWithFloat: points[0].x]];
				[pathArray addObject:[NSNumber numberWithFloat: points[0].y]];
					break;
			default:
				break;
		}
	}
	free(points);
	[array addObject: pathArray];
	
	
	NSColor *tmpColor = color;
	if (![[tmpColor colorSpaceName] isEqualToString:@"NSCalibratedRGBColorSpace"])
		if ((tmpColor = [tmpColor colorUsingColorSpaceName: @"NSCalibratedRGBColorSpace"]) == nil)
			tmpColor = [[NSColor blackColor] colorUsingColorSpaceName: @"NSCalibratedRGBColorSpace"];
    [array addObject: [NSNumber numberWithFloat:[tmpColor redComponent]]];
    [array addObject: [NSNumber numberWithFloat:[tmpColor greenComponent]]];
    [array addObject: [NSNumber numberWithFloat:[tmpColor blueComponent]]];
	[array addObject: [NSNumber numberWithFloat:thickness]];

	if (startRect)
        [array addObject:[NSNumber numberWithInt: [startRect rectID]]];
    else
        [array addObject:[NSNumber numberWithInt:0]];
	
    if (endRect)
        [array addObject:[NSNumber numberWithInt: [endRect rectID]]];
    else
        [array addObject:[NSNumber numberWithInt:0]];
	
	[array addObject:[NSNumber numberWithInt:arrow]];
	
	return [array autorelease];
}

- (void) exportFlash: (NSMutableString *) data;
{
	NSPoint *points = malloc(3*sizeof(NSPoint));
	NSBezierPathElement element = [path elementAtIndex:0 associatedPoints:points];
	//float startX = points[0].x;
	//float startY = points[0].y;
	
	int i, c = [path elementCount];
	for (i=1;i<c;i++) 
	{
		element=[path elementAtIndex:i associatedPoints:points];
		switch (element) 
		{
			case NSCurveToBezierPathElement:				
				[data appendFormat: @"<curve x=\"%.0f\" y=\"%.0f\" c1X=\"%.0f\" c1Y=\"%.0f\" c2X=\"%.0f\" c2Y=\"%.0f\" />\n", points[2].x, points[2].y * -1, points[0].x, points[0].y  * -1, points[1].x, points[1].y  * -1];
				break;
			case NSLineToBezierPathElement:
			case NSMoveToBezierPathElement:
				[data appendFormat: @"<line x=\"%.0f\" y=\"%.0f\" />\n", points[0].x, points[0].y * -1];
				break;
			default:
				break;
		}
	}
	free(points);	
}


- (id)initWithContentsOfArray:(NSArray *) array forDocument: (OCEDocument *) newDocument
{
	NSArray *pathArray = [array objectAtIndex:2];
	int kind, i = 3;
	int c = [pathArray count];
	path = [[NSBezierPath bezierPath] retain];
    [path moveToPoint:NSMakePoint([[pathArray objectAtIndex:1] floatValue],[[pathArray objectAtIndex:2] floatValue])];

	while (i<c)
	{
		kind = [[pathArray objectAtIndex:i] intValue];
		
		if (kind == 0)
		{
			[self addCurveToPoint: NSMakePoint([[pathArray objectAtIndex:i+5] floatValue],
											   [[pathArray objectAtIndex:i+6] floatValue])
					controlPoint1: NSMakePoint([[pathArray objectAtIndex:i+1] floatValue],
											   [[pathArray objectAtIndex:i+2] floatValue])
					controlPoint2: NSMakePoint([[pathArray objectAtIndex:i+3] floatValue],
											   [[pathArray objectAtIndex:i+4] floatValue])];
			i+=7;
		}
		else
		{
			[self addLineToPoint: NSMakePoint([[pathArray objectAtIndex:i+1] floatValue],
											  [[pathArray objectAtIndex:i+2] floatValue])];
			i+=3;
		}
	}
	
	[color release];
    color = [[NSColor colorWithCalibratedRed: [[array objectAtIndex:3] floatValue]
											 green: [[array objectAtIndex:4] floatValue]
											  blue: [[array objectAtIndex:5] floatValue] 
											 alpha: 1.0] retain];
	[self setThickness: [[array objectAtIndex:6] floatValue]];
	startRectID = [[array objectAtIndex:7] intValue];
	endRectID = [[array objectAtIndex:8] intValue];
	arrow = [[array objectAtIndex:9] intValue];
	
    return [super initWithContentsOfArray: array forDocument: newDocument];
}

- (void) finishLoadingObject: (OCEDocument *) document
{
    OCERectObject *object = nil;
    if (startRectID != 0)
    {
        object = [document rectObjectFromID:startRectID];
        if (object)
            [self setAssociatedStartRect:object];
    }
	
    object = nil;
    if (endRectID != 0)
    {
        object = [document rectObjectFromID:endRectID];
        if (object)
            [self setAssociatedEndRect:object];
    }
}


// ----------------------------------------------------------------------
// Associated Rects
// ----------------------------------------------------------------------
- (OCERectObject *) associatedStartRect
{
	return startRect;
}

- (void) setAssociatedStartRect: (OCERectObject *) rectObject
{
	startRect = rectObject;	
	[rectObject addAssociatedPath:self];
	[self adaptPathToRects];
	[[undoManager prepareWithInvocationTarget:self] unregisterAssociatedStartRect];
}

- (OCERectObject *) associatedEndRect
{
	return endRect;
}

- (void) setAssociatedEndRect: (OCERectObject *) rectObject
{
	endRect = rectObject;
	[rectObject addAssociatedPath:self];
	[self adaptPathToRects];
	[[undoManager prepareWithInvocationTarget:self] unregisterAssociatedEndRect];
}

- (void) unregisterAssociatedStartRect
{
	[[undoManager prepareWithInvocationTarget:self] setAssociatedStartRect: startRect];
	[startRect removeAssociatedPath:self];
	startRect = nil;
}

- (void) unregisterAssociatedEndRect
{
	[[undoManager prepareWithInvocationTarget:self] setAssociatedEndRect: endRect];
	[endRect removeAssociatedPath:self];
	endRect = nil;
}

- (void) unregisterAssociatedRects
{
	[self unregisterAssociatedStartRect];	
	[self unregisterAssociatedEndRect];	

}

- (void) unregisterAssociatedRect: (OCERectObject *) rectObject
{
	if (startRect==rectObject)
		[self unregisterAssociatedStartRect];	
	
	if (endRect==rectObject) 
		[self unregisterAssociatedEndRect];	
}

- (void) adaptPathToRects
{
	if ((startRect) && (endRect) && ([path elementCount] == 2))
	{
		NSArray *array = [[startRect path] intersectionsWithPath:[endRect path]];
		if ([array count])
		{
			[self setStartPoint:[startRect center]];
			[self setEndPoint:[endRect center]];
			return;
		}
	}
	
	if (startRect)
	{
		[self setStartPoint:[startRect center]];
		NSArray *array = [path intersectionsWithPath:[startRect path]];
		int count;
		if(count = [array count]) {
			while (count--) {
				NSPoint intersectPoint = [path getPointForPosition:[[array objectAtIndex:count] floatValue] 
														 andOffset:0];
				if ([path segmentHitByPoint:intersectPoint] == 1) {// segment 0 == moveToElement
					[self setStartPoint:intersectPoint];
					break;
				}
			}
		}		
	}
	if (endRect)
	{
		[self setEndPoint:[endRect center]];
		NSArray *array = [path intersectionsWithPath:[endRect path]];
		int count;
		if(count = [array count]) {
			while (count--) {
				NSPoint intersectPoint = [path getPointForPosition:[[array objectAtIndex:count] floatValue]
														 andOffset: 0];
				if ([path segmentHitByPoint:intersectPoint] == [path elementCount]-1) {
					[self setEndPoint:intersectPoint];
					break;
				}
			}
		}   
	}
}

- (void) removedFromCanvas: (OCEDocument *) document
{
	[self unregisterAssociatedRects];
}


@end
